import {LocationQueryValue} from "vue-router"

interface HttpResponse<T = unknown> {
  data: T;
  headers: Record<string, string>;
}

export class PaginatedList<T> {

  data: T[] = []
  totalCount = 0
  offset = 0
  limit = 0

  readonly loading: boolean

  constructor(loading = false) {
    this.loading = loading
  }

  /**
   * this 뒤에 list 의 항목들을 추가한다.
   *
   * - totalCount 가 새로운 list 의 totalCount 로 변경된다.
   * - limit 이 list.length 만큼 증가한다.
   *
   * @return this 객체
   */
  append(list: PaginatedList<T>): PaginatedList<T> {
    list.data.forEach((item) => this.data.push(item))
    this.totalCount = list.totalCount
    // append 의 경우 offset 은 변경없고, limit 이 늘어난다.
    this.limit += list.data.length
    return this
  }

  /**
   * this 앞에 list 의 항목들을 추가한다.
   *
   * - totalCount 가 새로운 list 의 totalCount 로 변경된다.
   * - offset 이 list.length 만큼 줄어든다.
   * - limit 이 list.length 만큼 증가한다.
   *
   * @return this 객체
   */
  prepend(list: PaginatedList<T>): PaginatedList<T> {
    list.data.forEach((item) => this.data.unshift(item))
    this.totalCount = list.totalCount
    // prepend 의 경우 offset 이 줄어들고, limit 이 늘어난다.
    this.offset -= list.data.length
    this.limit += list.limit
    return this
  }

  /**
   * PaginatedList 의 모든 데이터를 삭제하고,
   * totalCount=0, offset=0, limit=0 으로 변경한다.
   */
  clear(): void {
    this.data = []
    this.totalCount = 0
    this.offset = 0
    this.limit = 0
  }

  /**
   * PaginatedList 가 로딩중인지 확인한다.
   * `PaginatedList.loading()` 으로 생성한 것만이 로딩중인 PaginatedList 이다.
   */
  isLoading(): boolean {
    return this.loading
  }

  /**
   * @return 로딩된 데이터가 비어있으면 true
   */
  isEmpty(): boolean {
    return (this.data.length === 0)
  }

  static of<T>(data: T[] = [], totalCount: number | undefined = undefined, offset = 0, limit: number | undefined = undefined): PaginatedList<T> {
    const paginatedList = new PaginatedList<T>()
    paginatedList.data = data ? data.slice() : []
    paginatedList.totalCount = Math.max(0, totalCount || paginatedList.data.length)
    paginatedList.offset = Math.max(0, offset)
    paginatedList.limit = Math.max(0, limit || paginatedList.data.length)
    return paginatedList
  }

  /**
   * `AxiosResponse` 등의 http response 로부터 `PaginatedList` 를 생성한다.
   */
  static fromAxiosResponse<T, D = T>(response: HttpResponse<T[]>, converter?: (data: T) => D): PaginatedList<D> {
    const actualConverter = converter ? converter : ((item: T) => item as unknown as D)
    const data = (response.data && response.data.length) ? response.data.map((item) => actualConverter(item)) : [] as D[]
    return PaginatedList.of(data,
        Number(response.headers['x-total-count']) || undefined,
        Number(response.headers['x-offset']) || 0,
        Number(response.headers['x-limit'] || undefined))
  }

  /**
   * 로딩중을 나타내는 PaginatedList 를 생성한다.
   */
  static loading<T>(): PaginatedList<T> {
    return new PaginatedList<T>(true)
  }
}

/**
 * PaginatedList 를 가져오기 위한 요청 정보
 */
export type PaginatedListRequest = {
  offset: number
  limit: number
}

/**
 * PaginatedList 로드 정보
 */
export type PaginatedListLoadInfo = {
  totalCount: number,
  offset: number,
  limit: number
}

/**
 * 목록 검색 조건
 */
export class ListCondition {

  readonly values: string[]
  readonly defaultValue: string

  constructor(values: string[], defaultValue?: string) {
    this.values = values
    this.defaultValue = defaultValue || values[0]
  }

  toSelectItems(captionProvider?: (value?: string) => string | undefined): { value: string, caption: string }[] {
    return this.values.map(value => ({value, caption: captionProvider?.(value) || value}))
  }

  fromRouterQuery(queryValue: LocationQueryValue | LocationQueryValue[]): string {
    const value = String(queryValue)
    return (value && this.values.includes(value)) ? value : this.defaultValue
  }
}