import {App, Component, inject, onBeforeUnmount, onMounted} from "vue"
import {DynamicModalOptions, VueFinalModalProperty} from "vue-final-modal"
import {RouteLocationNormalized, Router} from "vue-router"

type ModalProps = { [key in string]: unknown }

type ClosableModal = {
  props: { name: string }
  toggle: () => void
  params: { value: { name: string } }
}

export type HistoryModalOptions = DynamicModalOptions & {
  // 화면을 다시 로드했을 때 historyModal 을 표시할지 여부. 디폴트는 false
  showOnRefresh?: boolean
}

// DynamicModalOptions 객체로부터 modal name 을 추출한다.
const getModalName = (modal: DynamicModalOptions): string => {
  const name = (typeof modal.component === 'string') ? modal.component : modal.component?.name || ''
  if (!name) throw Error(`Component name is missing. component = ${modal.component}`)
  return name
}

const getComponentName = (component: string | Component): string => {
  const name = (typeof component === 'string') ? component : component.name || ''
  if (!name) throw Error(`Component name is missing. component = ${component}`)
  return name
}

export class HistoryModalManager {

  private readonly historyModals: Map<string, HistoryModal[]> = new Map()

  private unregisterRouterBeforeEach?: () => void

  constructor(public readonly modal: VueFinalModalProperty,
              public readonly router: Router) {
    this.router.isReady().then(() => this.initialize())
  }

  async initialize(): Promise<void> {
    // router 의 beforeEach() 에서 historyModal 을 체크하도록 등록한다.
    this.unregisterRouterBeforeEach = this.router.beforeEach((to) => this.handleRouterBeforeEach(to))
    await this.synchronizeCurrent()
  }

  destroy(): void {
    // router 의 beforeEach() 등록을 해제한다.
    this.unregisterRouterBeforeEach?.()
  }

  async registerHistoryModal(modal: HistoryModal): Promise<void> {
    const historyModals = this.historyModals.get(this.getRouteName()) || []
    historyModals.push(modal)
    this.historyModals.set(this.getRouteName(), historyModals)

    // 방금 등록한 historyModal 이 현재 route hash 에 존재할 수도 있으므로, 여기서 한 번 동기화해야 한다.
    await this.synchronizeCurrent()
  }

  unregisterHistoryModal(modal: HistoryModal): void {
    const historyModals = this.historyModals.get(this.getRouteName()) || []
    const index = historyModals.indexOf(modal)
    if (index >= 0) historyModals.splice(index, 1)
    this.historyModals.set(this.getRouteName(), historyModals)
    if (this.historyModals.get(this.getRouteName())?.length === 0)
      this.historyModals.delete(this.getRouteName())
  }

  private getRouteName(): string {
    const route = this.router.currentRoute.value
    return route.name ? String(route.name) : ''
  }

  private async handleRouterBeforeEach(to: RouteLocationNormalized): Promise<void> {
    // url 이 변경될 때마다 hash 에 historyModal name 이 존재하는지 확인한다.
    await this.synchronizeModals(this.getModalNamesInHash(to.hash))
  }

  private getModalNamesInHash(hash = ''): string[] {
    return hash.startsWith('#') ? hash.substring(1).split(',') : []
  }

  // modalNames 에 해당하는 HistoryModalOptions 객체를 찾는다.
  private findModals(modalNames: string[]): HistoryModalOptions[] {
    const historyModals = this.historyModals.get(this.getRouteName()) || []

    return historyModals
        .flatMap(modal => modal.findRegisteredModals(modalNames))
  }

  // modalName 을 갖고 있는 HistoryModal 객체를 찾는다.
  private findHistoryModalHaving(modalName: string): HistoryModal | undefined {
    const modalsInGlobal = this.historyModals.get('') || []
    const modalsInRoute = this.historyModals.get(this.getRouteName()) || []
    const historyModals: HistoryModal[] = [...modalsInGlobal, ...modalsInRoute]

    return historyModals.find(modal => modal.hasModal(modalName))
  }

  /**
   * history modal 을 표시한다.
   * @param component modal 컴포넌트 이름 또는 객체
   * @param props 모달에 필요한 props
   */
  public async show(component: string | Component, props?: ModalProps): Promise<void> {
    const name = getComponentName(component)
    const historyModal = this.findHistoryModalHaving(name)
    const modal = historyModal?.getModal(name)
    if (historyModal && modal) {
      // modal 을 표시한다.
      this.internalShowModal(modal, name, props)

      // url 에 hash 를 추가한다.
      const route = this.router.currentRoute.value
      const hash = route.hash
      const newHash = (hash.length > 0) ? `${hash},${name}` : `#${name}`
      await this.router.push({query: route.query, hash: newHash})

      // name 을 갖고 있는 HistoryModal 에 modal 이 열린 순간의 fullPath 를 기록한다.
      historyModal.setFullPathOnShow(name, {
        before: route.fullPath,
        after: this.router.currentRoute.value.fullPath
      })
    }
  }

  private internalShowModal(modal: HistoryModalOptions, modalName: string, props?: ModalProps): void {
    const _modal = {...modal, bind: {...modal.bind, ...props}}
    this.modal.show(_modal, {name: modalName})
  }

  /**
   * history modal 을 close 한다.
   * @param component modal 컴포넌트 이름 또는 객체
   */
  public async close(component: string | Component): Promise<void> {
    const name = getComponentName(component)
    const fullPathOnShow = this.findHistoryModalHaving(name)?.getFullPathOnShow(name)
    await this.internalCloseModal(name, fullPathOnShow)
  }

  /**
   * 화면에 표시된 모든 modal 을 닫는다.
   * history modal 이 떠있는 경우, hash 도 제거된다.
   */
  public async closeAllModals(): Promise<void> {
    const route = this.router.currentRoute.value
    const modalNames = this.getModalNamesInHash(route.hash)
    for (const modalName of modalNames) {
      await this.close(modalName)
    }
  }

  async internalCloseModal(modalName: string, fullPathOnShow?: FullPathOnShow): Promise<void> {
    setTimeout(async () => {
      // modal 이 닫혔는데 hash 에 남아있는 경우, 제거한다.
      const route = this.router.currentRoute.value
      const names = this.getModalNamesInHash(route.hash)
      if (names.includes(modalName)) {

        if (fullPathOnShow?.after === route.fullPath) {
          // 현재 fullPath 가 modal 이 열린 순간의 fullPath 와 같다면,
          // 직전에 show() 를 통해 modal 을 연 상태이다.
          // router.back() 한다.
          this.router.back()
          // back() 이 완료될 때까지 기다린다.
          await this.waitFullPathChange(fullPathOnShow?.before)

        } else {
          // 그렇지 않으면 router.replace() 로 URL 을 변경한다.
          names.splice(names.indexOf(modalName), 1)
          const newHash = (names.length === 0) ? undefined : `#${names.join(',')}`
          await this.router.replace({query: route.query, hash: newHash})
        }
      }
    })
  }

  private async waitFullPathChange(fullPath?: string): Promise<void> {
    await new Promise((resolve) => {
      let maxTry = 10 // 최대 10번 확인
      const waitOneTime = () => {
        if ((maxTry <= 0) || (fullPath === undefined) || (this.router.currentRoute.value?.fullPath === fullPath)) {
          resolve(undefined)
        } else {
          maxTry--
          window.setTimeout(waitOneTime, 100) // 100ms 마다 확인
        }
      }
      waitOneTime()
    })
  }

  private async synchronizeCurrent(): Promise<void> {
    const hash = this.router.currentRoute?.value?.hash
    const modalNamesInHash = this.getModalNamesInHash(hash)
    await this.synchronizeModals(modalNamesInHash)
  }

  private async synchronizeModals(modalNamesInHash: string[]): Promise<void> {
    const historyModals = this.historyModals.get(this.getRouteName()) || []

    // 각 HistoryModal.modalMap 에 등록된 modal 의 이름
    const registeredModalNames = historyModals.flatMap(modal => modal.getRegisteredModalNames())

    // 현재 URL hash 에 없다면, 닫는다. (historyModal, 일반 modal 포함)
    this.modal.openedModals
        .filter(open => {
          const name = (open as unknown as ClosableModal).params?.value?.name
          return !modalNamesInHash.includes(name)
        })
        .forEach(toHide => (toHide as unknown as ClosableModal).toggle())

    // this.modalMap 에 등록된 modal 이고, 현재 URL hash 에 있는데 열려 있지 않다면,
    // - showOnRefresh == true 이면 연다.
    // - 그렇지 않으면 hash 를 없앤다.
    const openedNames = this.modal.openedModals
        .map(open => (open as unknown as ClosableModal).params?.value?.name)
        .filter(name => !!name)
    const modalsInHash = this.findModals(modalNamesInHash)
    modalsInHash
        .filter(modal => {
          const name = getModalName(modal)
          return registeredModalNames.includes(name) && !openedNames.includes(name)
        })
        .forEach((toOpen) => {
          if (toOpen.showOnRefresh) {
            this.internalShowModal(toOpen, getModalName(toOpen))
          } else {
            this.internalCloseModal(getModalName(toOpen), undefined)
          }
        })
  }
}

type FullPathOnShow = {
  before: string
  after: string
}

export class HistoryModal {

  private readonly manager

  private modalMap: Map<string, HistoryModalOptions> = new Map()
  private fullPathOnShow: Map<string, FullPathOnShow> = new Map() // this.show() 할 때의 route.fullPath

  constructor(manager: HistoryModalManager) {
    this.manager = manager
  }

  async initialize(): Promise<void> {
    await this.manager.registerHistoryModal(this)
  }

  destroy(): void {
    this.manager.unregisterHistoryModal(this)
  }

  register(modal: HistoryModalOptions): void {
    if (!modal.component) throw Error('component is required in DynamicModalOptions')
    const modalName = getModalName(modal)
    modal.on = {...modal.on, closed: async () => await this.closedHandler(modalName)}
    this.modalMap.set(modalName, modal)
  }

  getRegisteredModalNames(): string[] {
    return Array.from(this.modalMap.keys())
  }

  findRegisteredModals(modalNames: string[]): HistoryModalOptions[] {
    return Array.from(this.modalMap.entries())
        .filter((entry) => modalNames.includes(entry[0]))
        .map((entry) => entry[1])
  }

  hasModal(modalName: string): boolean {
    return this.modalMap.has(modalName)
  }

  getModal(modalName: string): HistoryModalOptions | undefined {
    return this.modalMap.get(modalName)
  }

  setFullPathOnShow(modalName: string, fullPath: FullPathOnShow): void {
    this.fullPathOnShow.set(modalName, fullPath)
  }

  getFullPathOnShow(modalName: string): FullPathOnShow | undefined {
    return this.fullPathOnShow.get(modalName)
  }

  private async closedHandler(modalName: string): Promise<void> {
    const fullPathOnShow = this.fullPathOnShow.get(modalName)
    await this.manager.internalCloseModal(modalName, fullPathOnShow)
    // 닫혔으므로, fullPathOnModalOpen 에서 제거
    this.fullPathOnShow.delete(modalName)
  }
}

const historyModalKey = Symbol('HistoryModalPluginKey')

/**
 * 반드시 `setup()` 함수 내에서 호출되어야 한다.
 * @param modals
 */
export const useHistoryModal = (...modals: HistoryModalOptions[]): HistoryModalManager => {

  const manager = inject<HistoryModalManager>(historyModalKey)
  if (!manager) {
    throw new Error('HistoryModalPlugin is not installed!')
  }

  if (modals.length > 0) {
    const historyModal = new HistoryModal(manager)
    modals.forEach(modal => historyModal.register(modal))

    onMounted(async () => await historyModal.initialize())
    onBeforeUnmount(() => historyModal.destroy())
  }

  return manager
}

export type HistoryModalPluginOption = {
  modal: VueFinalModalProperty
  router: Router
}

/**
 * ```ts
 * createApp(GuideMain)
 *     .use(router)
 *     .use(ModalPlugin)
 *     .use(HistoryModalPlugin, {router, modal: ModalPlugin.modal})
 * ```
 */
export const HistoryModalPlugin = {

  install(app: App, option?: HistoryModalPluginOption): void {
    if (!option) {
      throw new Error("option (HistoryModalPluginOptions) is missing!")
    }
    if (!option.modal) {
      throw new Error('option.modal is missing!')
    }
    if (!option.router) {
      throw new Error('option.router is missing!')
    }

    const manager = new HistoryModalManager(option.modal, option.router)
    app.provide(historyModalKey, manager)
  }
}